/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               | 
+--------------------------------------------------------------------*/
// $Id: FtnLexer.cc,v 1.8 1999/10/14 11:20:44 fgg Exp $

#include <string.h>
#include <ctype.h>
#include "Octave/Fortran.h"

static char* putLineNumber (char*, int) ;


/*--------------------------------------------------------+
 |                                                        |
 |      Static structures for recognising keywords        |
 |            and Convex-specific directives              |
 |                                                        |
 +--------------------------------------------------------*/


// Pointers to start and end locations in the keyword table
// for each set of keywords starting with a different letter.

static struct Keylist *keystart[26], *keyend[26];

// Pointers to start and end locations in the directive table
// for each set of directives starting with a different letter.

static struct Keylist *dirstart[26], *dirend[26];

// Table of alpha-sorted recognised statement keywords
// and directives for Convex fortran.

static struct Keylist keys[] = {
        {"accept", ACCEPT},
        {"allocatable", ALLOCATABLE},
        {"allocate", ALLOCATE},
	{"assignment", ASSIGNMENT},
	{"assign", ASSIGN},
	{"backspace", BACKSPACE},
	{"blockdata", BLOCKDATA},
	{"byte", BYTE},
	{"call", CALL},
	{"character", CHARACTER},
        {"close", CLOSE},
	{"common", COMMON},
	{"complex", COMPLEX},
	{"continue", CONTINUE},
	{"data", DATA},
        {"deallocate", DEALLOCATE},
	{"decode", DECODE },
	{"dimension", DIMENSION},
	{"doubleprecision", DOUBLEPRECISION},
	{"doublecomplex", DOUBLECOMPLEX},
	{"else", ELSE},
	{"encode", ENCODE },
	{"enddo", ENDDO},
	{"endfile", ENDFILE},
	{"endif", ENDIF},
	{"end", ENDUNIT},
	{"entry", ENTRY},
	{"equivalence", EQUIVALENCE},
	{"external", EXTERNAL},
	{"find", FIND},
	{"format", FORMAT_TOKEN},
	{"function", FUNCTION},
	{"goto", PLAINGOTO},
	{"implicitnone", IMPLICITNONE},
	{"implicit", IMPLICIT},
	{"include", INCLUDE},
	{"inquire", INQUIRE},
	{"integer", INTEGER},
	{"intrinsic", INTRINSIC},
	{"logical", LOGICAL},
	{"namelist", NAMELIST},
	{"open", OPEN},
        {"options", OPTIONS},
	{"parameter", PARAMETER},
	{"pause", PAUSE},
        {"pointer", POINTER},
	{"print", PRINT},
	{"program", PROGRAM},
	{"read", READ},
	{"real", REAL},
	{"record", RECORD},
	{"return", RETURN},
	{"rewind", REWIND},
	{"save", SAVE},
	{"stop", STOP},
	{"subroutine", SUBROUTINE},
	{"taskcommon", TASK_COMMON},
	{"then", THEN},
	{"to", TO},
	{"type", TYPE},
	{"write", WRITE},
	{0, 0}
};

static struct Keylist dirs[] = {
        {"assign_lock", ASSIGN_LOCK},
        {"begin_order", BEGIN_ORDER},
	{"begin_section", BEGIN_SECTION},
	{"begin_tasks", BEGIN_TASKS},
	{"end_order", END_ORDER},
	{"end_section", END_SECTION},
	{"end_tasks", END_TASKS},
	{"force_parallel_ext", FORCE_PARALLEL_EXT},
	{"force_parallel", FORCE_PARALLEL},
	{"force_vector", FORCE_VECTOR},
	{"free_lock", FREE_LOCK},
	{"max_trips", MAX_TRIPS},
	{"next_task", NEXT_TASK},
	{"no_parallel", NO_PARALLEL},
	{"no_peel", NO_PEEL},
	{"no_promote_test", NO_PROMOTE_TEST},
	{"no_recurrence", NO_RECURRENCE},
	{"no_side_effects", NO_SIDE_EFFECTS},
	{"no_vector", NO_VECTOR},
	{"peel_all", PEEL_ALL},
        {"peel", PEEL},
	{"promote_test_all", PROMOTE_TEST_ALL},
	{"promote_test", PROMOTE_TEST},
        {"prefer_parallel_ext", PREFER_PARALLEL_EXT},
        {"prefer_parallel", PREFER_PARALLEL},
        {"prefer_vector", PREFER_VECTOR},
        {"pstrip", PSTRIP},
        {"row_wise", ROW_WISE},
        {"scalar", SCALAR},
	{"select", SELECT},
	{"synch_parallel", SYNCH_PARALLEL},
        {"unroll", UNROLL},
        {"vstrip", VSTRIP},
	{0, 0}
};


void FortranLexer::initkeys(struct Keylist *names, 
			    struct Keylist *keystart[26],
			    struct Keylist *keyend[26])
{
  struct Keylist *p ;
  register int i, j ;
  
  for (i = 0; i < 26; ++i)
    keystart[i] = 0 ;
  
  for (p = names; p->keyname; ++p) {
    j = p->keyname[0] - 'a';
    if (keystart[j] == 0)
      keystart[j] = p ;
    keyend[j] = p + 1 ;
  }
}

const char* FortranLexer::matchkeyword (const char* str, int* token)
{
  struct Keylist *pk, *pend ;
  int k = tolower(*str) - 'a' ;

  if (pk = keystart[k])
    for (pend = keyend[k]; pk != pend; ++pk) {
      const char* i = pk->keyname;
      const char* j = str ;

      while (*++i == tolower(*++j) && *i != '\0' && *j != '\0') ;

      if (*i == '\0')
	{
	  *token = pk->keyval ;
	  return j ;
	}
    }
  *token = UNKNOWN_TOK ;
  return str ;
}

const char* FortranLexer::matchdirective (const char* str, int* token)
{
  struct Keylist *pk, *pend ;
  int k = tolower(*str) - 'a' ;

  if (pk = dirstart[k])
    for (pend = dirend[k]; pk != pend; ++pk) {
      const char* i = pk->keyname;
      const char* j = str ;

      while (*++i == tolower(*++j) && *i != '\0' && *j != '\0') ;

      if (*i == '\0')
	{
	  *token = pk->keyval ;
	  return j ;
	}
    }
  *token = UNKNOWN_TOK ;
  return str ;
}

void FortranLexer::LexerInit()
{
  nchars = 0 ;
  ip_position = 1 ;
  pos_in_line = 0 ;
  in_dig_seq = 0 ;
  holl_count = dig_pos = dig_copied = 0 ;
  ip = rawBuf ;
  op = bp = eos = scanBuf ;
  end_of_scanBuf = &scanBuf[buf_size-1] ;
  cwp = crp = eoc = cmntBuf ;
  lines_scanned = 0 ;
  next_stmt_line = 1 ;
  is_blank = 1 ;
  is_continuation = 0 ;
  scan_state = old_state = IN_PREFIX ;
  initkeys (keys, keystart, keyend) ;
  initkeys (dirs, dirstart, dirend) ;
}

int FortranLexer::is_data_statement ()
{
  // match the start of statement with: 
  //
  //        " "[0-9]*[dD][aA]

  // Assume that enough of the statement has been read to
  // make a determination possible, if not then we return
  // false always.

  char *b = bp ; // scan from start of current statement

  if (*b++ != ' ') return 0 ;
  while (*b >= '0' && *b <= '9' && b != op) 
    b++ ;
  if (( (b+1) < op) &&
      (*b == 'd' || *b == 'D') && 
      (b[1] == 'a' || b[1] == 'A'))
    return 1 ;
  else
    return 0 ;
}

/*-----------------------------------------------------------

  FortranLexer::ScanStatement

  This routine pre-scans a complete line of Fortran source text
  into a private FortranLexer buffer, removing white space except when
  inside character strings or holleriths, and eliminating
  continuation line sequences. It also discards trailing
  comment characters, from a "!" character to the end of the
  line, and ignores comment lines that are positioned between
  either an initial line or a continuation line and another
  continuation line.

  Holleriths are tricky: essentially every time we start a
  sequence of integers we could be starting a hollerith,
  provided we don't start looking until we are past
  columns 1-6 in the input stream, or we have seen a leading tab.
  We note the starting position of the first digit
  in the output buffer. If we see a "H" or an "L" character
  immediately after the last digit (and there have been no
  characters that are not either white space or digits in
  between) we enter "hollerith" mode. We first compute
  the hollerith length, and then replace the already scanned
  digits with a \' character.
  Whilst we are in this mode all characters up to the hollerith
  length or the end of line (which ever comes sooner) are
  copied to the output buffer without deleting white space,
  taking care to explode all \' characters into \'\' sequences.
  When the hollerith is exhausted, we emit another "\'" character
  to close off the apostrophe-delimited string.

  Thus the FortranLexer object keeps the following state information;

   1. the start position of the last unbroken digit sequence.
      in the output buffer.
   2. a flag to denote if we are in a digit sequence.
   3. the hollerith count (which gets decremented as its read)
      this doubles-up as the "hollerith mode" flag.

 -----------------------------------------------------------*/

void FortranLexer::ScanStatement ()
{
  // Copies characters from rawBuf to scanBuf until a complete
  // Fortran statement or comment line has been scanned.
  //
  // Follows the lexical rules for crunching white space outside
  // strings and comments, and for converting holleriths to
  // apostrophe-delimited strings.
  //
  // When the number of unread characters in rawBuf falls below
  // 8 we try to top it up from the input stream.
  //
  // We terminate the copy loop when rawBuf is completely empty,
  // or when a complete line has been transfered to the scan
  // buffer.

//#define DEBUG_LEXER 1

#define BUFFER_SLACK 72
#define F77_LINE_LENGTH 72
#define CONSUME(_n_) ip += (_n_); nchars -= (_n_); ip_position += (_n_);
#define REWIND_BUFFER(_b_,_n_) (_b_) -= (_n_)
#define TRANSMIT(_c_,_b_) *(_b_)++ = (_c_);

  checkBuffers(__LINE__) ;
  if (eos == bp)
    {
      eos = op ;
    }

  checkBuffers(__LINE__) ;
  while (1)
    {
      // First ensure we have sufficient characters in
      // the raw input buffer.

      checkBuffers(__LINE__);
/* was before fgg changed on 2000/12/06
   otherwise there is a problem if the end of buffer
   happens while we remove spaces at the end
      if (nchars < 8)
*/
      if (nchars < F77_LINE_LENGTH+8)
	{
	  // Move unused input characters to the start
	  // of `rawBuf`, and call yyin->read() to get
	  // further input characters into the next free
	  // character position after the un-processed
	  // text. We never move more than 7 characters.
	  //
	  // Read up to `buf_size-nchars` characters into `rawBuf`
	  // add to `nchars` the number of characters read.
	  // set `ip` to start of rawBuf
	  
	  if (!(yyin->eof() || yyin->fail()))
	    {
	      char *wp = rawBuf ;
	      char *rp = ip ;
	      int  ccnt = nchars ;

	      while (ccnt)
		{ 
		  *wp++ = *ip++ ;
		  ccnt-- ;
		}

	      (void) yyin->read( wp, buf_size - nchars ) ;
	      nchars += yyin->gcount() ;
	      ip = rawBuf ;
	    }

	  // if no more characters, then end of input

	  if (!nchars)
	    {
	      eos = op ;
	      eoc = cwp ;
	      return ;
	    }
	}      
      checkBuffers(__LINE__);
      char nextch = *ip ;

      pos_in_line++;

      if (nextch=='\r')
      {
        //fgg added on 2000/12/08
        //to deal with Windows/DOS files
        //and not to crash.
        //Currently we just remove that character.
        //Later we can also make unparse in Windows/DOS standard.
        
        CONSUME(1);        
      }
      else if ((nextch!='\n') && 
               (pos_in_line>F77_LINE_LENGTH+1) && 
               (scan_state!=IN_COMMENT) &&
               (scan_state!=IN_EXCLUDE))
      {
        //fgg added on 2000/12/05
        //dummy - just ignore everything after 72 position
        //TBD - add this as a comment
        //may have problems if this comment is between continuation of lines

        CONSUME(1);
      }
      else if (nextch == '\n')
	{
	  /* On end of line, we must check for:-
	   *
	   *   (i) Blank line (remove it)
	   *
	   *  (ii) Completed non-blank line:
	   *        - if it was a continuation, continue scanning
	   *        - otherwise return if eos != bp
	   *
	   * (iii) Partially complete hollerith; which must be
	   *       properly terminated.
	   */

          pos_in_line=0;
	  
	  CONSUME(1) ;

	  switch (scan_state)
	    {
	    case IN_COMMENT:
              if ((cwp-cmntBuf+1000)<buf_size)
              {
                //fgg added on 2000/12/08
                //to deal with comments more then the buf_size.
                //Currently we just ignore characters if they can't go to buffer.
                //TBD
                TRANSMIT(nextch, cwp) ;
                eoc = cwp ;
              }
	      scan_state = IN_PREFIX ;
	      is_blank = 1 ;
	      if (op == bp)
		return ;
	      break ;

	    case IN_EXCLUDE:
	      scan_state = IN_STATEMENT ;
	    case IN_APOSTROPHE:
	    case IN_QUOTATION:
	    case IN_HOLLERITH:
	      if (scan_state == IN_HOLLERITH)
		{
		  TRANSMIT(HOLL_DELIM, op) ;  // provisional termination
		}
	    case IN_PREFIX:
	    case IN_STATEMENT:
            default: // fgg added on 2000/12/04
	      TRANSMIT(nextch, op) ;
	  
	      if (is_blank && !is_continuation)
		{
		  ++lines_scanned ;
		  op = eos ;
		  scan_state = IN_PREFIX ;
		}
	      else if (is_continuation && !is_blank)
		{ // end of a non-blank continuation line
		  old_state = scan_state ;
		  scan_state = IN_PREFIX ;
		  is_blank = 1 ;
		  is_continuation = 0 ;
		  eos = op ;
		  eoc = cwp ;
		}
	      else
		{ // end of a non-blank, non-continuation, non-comment line,
		  // or a blank continuation line.
		  old_state = scan_state ;
		  scan_state = IN_PREFIX ;
		  
		  if (is_blank)
		    {
		      op = eos ;
		      TRANSMIT('\n', op) ;
		      is_continuation = 0 ;
		    }

		  is_blank = 1 ;
		  if (eos != bp)
		    return ;
		  else
		    {
		      eos = op ;
		      eoc = cwp ;
		    }
		}
	    }
	}
      else
	switch (scan_state)
	  {
	  case IN_STATEMENT:
	    switch (nextch)
	      {
	      case '\'':
		in_dig_seq = is_blank = 0 ;
		TRANSMIT(nextch, op) ;
		CONSUME(1) ;
		scan_state = IN_APOSTROPHE ;
		break ;
		
	      case '"':
		in_dig_seq = is_blank = 0 ;
		TRANSMIT(nextch, op) ;
		CONSUME(1) ;
		scan_state = IN_QUOTATION ;
		break ;
		
	      case '!':
		in_dig_seq = 0 ;
		TRANSMIT(nextch, op) ;
		CONSUME(1) ;
		scan_state = IN_EXCLUDE ;
		break ;
		
	      case '0':
	      case '1':
	      case '2':
	      case '3':
	      case '4':
	      case '5':
	      case '6':
	      case '7':
	      case '8':
	      case '9':
		if ( (ip_position > 6) && !in_dig_seq )
		  {
		    in_dig_seq = 1 ;      // If we eventually find a hollerith
		    digit_start = op ;    // we revert ip_position to dig_pos
		    dig_pos = ip_position ;
		  }
		TRANSMIT(nextch, op) ;
		CONSUME(1) ;
		is_blank = 0 ;
		break ;
		
	      case 'H':
	      case 'h':
	      case 'L':
	      case 'l':
		// We may have found a hollerith. We can resolve this by
		// the following rule. A hollerith must be preceded by either:
		//
		//     (i). a punctuation mark in '/', '(', ',', '=', '.'
		// or
		//    (ii). a '*' character, provided this is a DATA statement.
		//
		// If this is not the case, then we simply clear in_dig_seq
		// without progressing the input pointer.
		if (in_dig_seq)
		  {
		    char p = digit_start[-1] ;
		    CONSUME(1) ;
		    is_blank = 0 ;
		    
		    if (p == '/' || p == '(' || p == ',' || p == '.' ||
			p == '=' || (p == '*' && is_data_statement()))
		      {
			// We have definitely found a hollerith, so compute
			// the hollerith count, reset the output buffer pointer
			// to the start of the numeric prefix, set position
			// to dig_pos, and then emit an opening \' character
			// to start the hollerith conversion.
			
			*op = 'H' ;
			holl_count = atoi (digit_start) ;
			op = digit_start ;
			ip_position = dig_pos ;
			TRANSMIT(HOLL_DELIM, op) ;
			scan_state = IN_HOLLERITH ;
			in_dig_seq = is_blank = 0 ;
		      }
		    else
		      {
			in_dig_seq = 0 ;
			TRANSMIT(nextch, op) ;
		      }
		  }
		else
		  {
		    is_blank = 0 ;
		    TRANSMIT(nextch, op) ;
		    CONSUME(1) ;
		  }
		break ;
		
	      case ';':
		in_dig_seq = 0 ;
		CONSUME(1) ;
		if (is_blank)
		  {
		    // Remove the current blank line, and
		    // continue scanning for a statement.
		    op = eos ;
		  }
		else
		  {
		    // This is the end of a statement.
		    // If (eos != bp) then we can be 
		    // sure that the statement in scanBuf
		    // from bp to eos is complete.
		    if (eos != bp)
		      {
			TRANSMIT('\n', op) ;
			return ;
		      }
		  }
		break ;
		
	      case ' ':
	      case '\t':
		// Normally we eliminate white space between characters at this
		// point, but we must be careful to leave in any such characters
		// that are needed to disambiguate certain character sequences.
		// e.g. '4 b' and '4b', which are INT_CONST IDENTIFIER in the
		// first case and OCTAL_CONST in the second.
		// if (in_dig_seq)
		//   TRANSMIT(nextch, op) ;
		CONSUME(1) ;
		in_dig_seq = 0 ;
		break ;
		
	      default:
		{
		  is_blank = in_dig_seq = 0 ;
		  TRANSMIT(nextch, op) ;
		  CONSUME(1) ;
		}
	      }
	    break ;
	    
	  case IN_COMMENT:
            CONSUME(1);        
            if ((cwp-cmntBuf+1000)<buf_size)
            {
              //fgg added on 2000/12/08
              //to deal with comments more then the buf_size.
              //Currently we just ignore characters if they can't go to buffer.
              //TBD
              TRANSMIT(nextch, cwp) ;
            }

	    break ;
	    
	  case IN_EXCLUDE:
	    TRANSMIT(nextch, op) ;
	    CONSUME(1) ;
	    break ;
	    
	  case IN_APOSTROPHE:
	    if (nextch == '\'')
	      scan_state = IN_STATEMENT ;
	    
	    TRANSMIT(nextch, op) ;
	    CONSUME(1) ;
            is_blank = 0; // FGG added on 2000/12/04
	    break ;
	    
	  case IN_QUOTATION:
	    if (nextch == '"')
	      scan_state = IN_STATEMENT ;

	    TRANSMIT(nextch, op) ;
	    CONSUME(1) ;
	    is_blank = 0 ;
	    break ;
	    
	  case IN_HOLLERITH:
	    is_blank = 0 ;

	    if (nextch == '\'')
	      TRANSMIT(nextch, op) ;  // explode apostrophe characters
	    
	    TRANSMIT(nextch, op) ;
	    CONSUME(1) ;
	    --holl_count ;
	    
	    if (holl_count == 0)
	      {
		// end of hollerith therefore terminate with \'
		TRANSMIT(HOLL_DELIM, op) ;
		scan_state = IN_STATEMENT ;
	      }
	    break ;
		
	  case IN_PREFIX:
	    ++lines_scanned ;
	    ip_position = 1 ;
	    is_continuation = 0 ;
	    is_blank = 1 ;
	    char nlower = tolower (nextch) ;
	    char n2 = 0 ;
	    char n6 = 0 ;
            bool shortline = true ;
            int ii = 0 ;
	    int ppos = 0 ;
	    char pfc ;

            /* FGG changes from 'ii < nchars' to 'ii < nchars-1' */
            while (ii < nchars-1 && ((pfc = ip[ii]) != '\n') && (shortline || ppos < 6))
              {
                if (pfc != ' ' && pfc != '\t')
                  shortline = false ;

		if (pfc == '\t') {
		  ppos += 6 ;
		  if (!n6)
		    n6 = ' ' ;
		}
		else
                {
		  ppos += 1 ;
                }

		switch (ppos)
		  {
		  case 2:
		    n2 = pfc ;
		    break ;
		  case 6:
		    n6 = pfc ;
		    break ;
		  default:
		    break ;
		  }
                ++ii ;
              }

	    if (shortline)
	      {
		CONSUME(ii+1)
	      }
	    else if (nlower == 'c' || nlower == 'd' || nextch == '!' || nextch == '*')
	      {
		// this line is a comment
               	is_blank = 0 ;
		scan_state = IN_COMMENT ;
                if ((cwp-cmntBuf+1000)<buf_size)
                {
                  //fgg added on 2000/12/08
                  //to deal with comments more then the buf_size.
                  //Currently we just ignore characters if they can't go to buffer.
                  //TBD
                  cwp = putLineNumber (cwp, lines_scanned) ;
                }
	      }
	    else if ( (nextch == ' ' && n6 != ' ' && n6 != '0') ||
		      (nextch == '\t' && n2 > '0' && n2 <= '9') )
	      {
		// this line is a continuation
		is_continuation = 1 ;
		if (nextch == '\t')
		  {
		    CONSUME(2) ;
		  }
		else
		  {
		    CONSUME(6) ;
		  }
		ip_position = 7 ;
                pos_in_line=7;
		scan_state = old_state ;
		if (eos == op)
		  REWIND_BUFFER(eos, 1) ;
		REWIND_BUFFER(op, 1) ;
		if (scan_state == IN_HOLLERITH)
		  {
		    if (eos == op)
		      REWIND_BUFFER(eos, 1) ;
		    REWIND_BUFFER(op, 1) ;
		  }
		checkBuffers(__LINE__) ;
	      }
	    else 
	      {
		op = putLineNumber (op, lines_scanned) ;
		TRANSMIT(' ', op) ;
		scan_state = IN_STATEMENT ;
		in_dig_seq = 0 ;
		if (nextch == '\t')
		  {
		    CONSUME(1) ;
		    ip_position = 7 ;
                    pos_in_line=7;
		  }
	      }
	    break ;
	  } // switch (scan_state)
      
      // Check for nearing end of scanBuf and re-position.
      
      checkBuffers(__LINE__);
      if (op >= (end_of_scanBuf - BUFFER_SLACK))
	{
	  int shift = bp - scanBuf ;
	  if (!shift)
	    LexerError ("lexical analysis buffer size exceeded\n") ;
	  else
	    {
	      int amount = op - bp ;
	      op = scanBuf ;
	      for (int i = amount; i; --i)
		*op++ = *bp++ ;
	      bp = scanBuf ;
	      eos -= shift ;
	      digit_start -= shift ;
	    }
	  checkBuffers(__LINE__);
	}
    }
}


void FortranLexer::checkBuffers (int id)
{
  // check for consistent state of scanBuf

  if (bp < scanBuf || eos < bp || op < eos ||
      end_of_scanBuf < op )
    {
      fprintf (stderr,"Buffer inconsistency (FtnLexer:%d) bp = %d, eos = %d, op = %d\n",
	       id, bp-scanBuf, eos-scanBuf, op-scanBuf) ;

      fprintf (stderr, "Buffer contents [bp..op-1]\n") ;
      char* t = bp ;
      while (t != op)
	fprintf (stderr, "%c", *t++) ;
      fprintf (stderr, "\n\n") ;
    }
}


int FortranLexer::LexerInput (char *buf, int max_size)
{
  int cmnt_size, stmt_size, i ;

  if (bp == eos && crp == eoc)
    ScanStatement () ;

  // Copy any statement text to buf

  // cout << "LexerInput:[" << bp-scanBuf << ", " << eos-scanBuf << "]" ;

  stmt_size = (eos - bp) > max_size ? max_size : eos - bp ;

  for (i = stmt_size; i; --i)
    {
      char cc = *bp ;
      // cout << cc ;
      *buf++ = *bp++ ;
    }
  // cout <<"\n" ;

  max_size -= stmt_size ;

  // Copy any trailing comment text to buf

  cmnt_size = (eoc - crp) > max_size ? max_size : eoc - crp ;
  for (i = cmnt_size ; i; --i)
    *buf++ = *crp++ ;
      
  // reset comment buffer when it is empty
  
  if (crp == cwp)
    crp = cwp = eoc = cmntBuf ;

  return cmnt_size + stmt_size ;
}

static char* putLineNumber (char* s, int l)
{
  char buf[20] ;
  char* bp = buf ;
  const int ten = 10 ;

  do
    {
      const int d = l/ten ;
      *bp++ = '0' + l - (d * ten) ;
      l = l/ten ;
    } while (l) ;

  while (bp != buf)
    {
      --bp ;
      *s++ = *bp ;
    }

  *s++ = ':' ;
  return s ;
}


// EVENTUALLY THIS WILL THROW AN EXCEPTION
// ..TO BE CAUGHT BY THE PARSER.. WHICH CALLS
//   THE LEXER WITHIN A TRY..CATCH BLOCK.

void FortranLexer::LexerError (const char* msg)
{
  cerr << msg ;
  // exit (EXIT_FAILURE) ;
}

