/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Source of test routines for library of set operations and set classes

#include <time.h>     // for benchmarking the set operations

#include "Octave/Base/Basic.h"
#include "Octave/Base/Sets.h"     // for library of Set classes

// this is a class definition for a contained class

class Tuple {
public:
  Tuple(int i) { ix = i ; }
  ~Tuple()     { }

  int index()  { return ix ; }
  void setIndex(int i) { ix = i ; }
  int variant() { return 0; }

private:
  int ix ;
  int a, b, c ;
} ;

// Now, to keep some compilers happy, we explicitly instantiate
// the class SetOf<Tuple>

template class SetOf<Tuple> ;

class TupleSet : public SetOf<Tuple> {
public:
  int index () { return id ; }
  int variant () { return 0 ; }
  void setIndex (int i) { id = i ; }
  TupleSet (int i) : SetOf<Tuple> () { id = i ; }

private:
  int id ;
} ;

void basic_regression()
{
  Universe<FiniteInt>* ut = new Universe<FiniteInt> () ;
  ut->make_member (1) ;
  ut->make_member (15) ;
  ut->make_member (43) ;
  ut->make_member (55) ;
  ut->make_member (58) ;
  ut->make_member (67) ;
  ut->make_member (77) ;
  ut->make_member (451) ;
  ut->make_member (1600) ;
  ut->make_member (2345) ;
  ut->make_member (2901) ;
  ut->make_member (10224) ;
  ut->make_member (16643) ;
  ut->make_member (654321) ;

  Set s1(ut) ;
  Set s2(ut) ;
  Set s3(ut) ;
  Set s4(ut) ;
  Set s5(ut) ;
  Set s6(ut) ;
  char *c1 ;
  int errs = 0 ;

  printf ("\nREGRESSION TESTS ON BASIC SETS\n\n") ;
  // Test 1: simple addition and deletion from a set

  printf (" TEST 1: add member, remove member, idempotency checks\n") ;
  printf ("  s1 = %s\n", c1 = s1.print(1, 0)) ;  //should be empty
  delete c1 ;
  s1 += 654321 ;
  printf ("  s1 = %s\n", c1 = s1.print(1, 0)) ;  //should be {654321}
  delete c1 ;
  s1 += 15 ;
  printf ("  s1 = %s\n", c1 = s1.print(1, 0)) ;  //should be {15, 654321}
  delete c1 ;
  s1 -= 654321 ;
  printf ("  s1 = %s\n", c1 = s1.print(1, 0)) ;  //should be {15}
  delete c1 ;
  s1 -= 15 ;
  printf ("  s1 = %s\n", c1 = s1.print(1, 0)) ;  //should be empty
  delete c1 ;

  // check if is_empty() returns true

  if (s1.is_empty())
    printf ("  'is_empty()' predicate functioning OK\n") ;
  else
    printf ("  'ERROR: 'is_empty()' predicate claims s1 is non-empty\n") ;

  // Test 2: as test 1, but with a complemented set
  printf (" TEST 2: addition and subtraction from a univeral set\n") ;
  Universe<FiniteInt>* uv = new Universe<FiniteInt> () ;
  for (int kk = 150; kk < 154; kk++)
    uv->make_member (kk) ;
  Set sp (uv) ;
  sp.empty() ;
  printf ("  sp = %s\n", c1 = s1.print(1, 0)) ;  //should be empty
  delete c1 ;
  sp.make_universal () ;     // s1 becomes U(150,153) - {}
  printf ("  sp = %s\n", c1 = sp.print(1, 0)) ;
  delete c1 ;
  sp -= 150 ;
  printf ("  sp = %s\n", c1 = sp.print(1, 0)) ;  // should be U-{150}
  delete c1 ;
  sp -= 151 ;
  sp -= 152 ;
  sp -= 153 ;
  printf (" The next line should be a fatal error...\n") ;
  sp -= 154 ;
  printf ("  sp = %s\n", c1 = sp.print(1, 0)) ;  //should be empty
  delete c1 ;
  if (sp.is_empty())
    printf ("  'is_empty()' predicate OK on complemented empty set\n") ;
  else {
    printf ("  ERROR: 'is_empty()' predicate claims s1 is non-empty\n") ;
    ++errs ;
  }
  printf ("  sp = %s\n", c1 = sp.print(1, 0)) ;  delete c1 ;

  // Test 3: Set Union operator...

  printf (" TEST 3: Set Union operator\n") ;
  s1.empty() ;
  s2.empty() ;
  s3.empty() ;
  s1 += 67 ;
  s1 += 451 ;
  printf ("  given s1 = %s", c1 = s1.print(1, 0)) ; delete c1 ;
  s2 += 77 ;
  s2 += 451 ;
  s2 += 10224 ;
  printf (", s2 = %s", c1 = s2.print(1,0)) ; delete c1 ;
  s3 += 16643 ;
  s3 += 1 ;
  s3 += 43 ;
  printf (", and s3 = %s\n", c1 = s3.print(1,0)) ; delete c1 ;
  s4 = s1 + (s2 + s3) ;
  printf ("  then s4 = s1 + s2 + s3 is  %s\n", c1 = s4.print(1,0)) ;delete c1 ;
  s5.empty() ;
  s5 += 1 ;
  s5 += 43 ;
  s5 += 67 ;
  s5 += 451 ;
  s5 += 77 ;
  s5 += 10224 ;
  s5 += 16643 ;
  if (s4 == s5)
    printf ("  set union operator OK\n") ;
  else {
    printf ("  ERROR: set union operator FAILED\n") ;
    ++errs ;
  }

  // TEST 4: Set intersection operator

  printf ("\n TEST 4: Set intersection operator\n") ;
  s1.empty() ;
  s2.empty() ;
  s3.empty() ;
  s1 += 55 ;
  s1 += 58 ;
  s1 += 2345 ;
  s1 += 2901 ;
  printf ("  given s1 = %s", c1 = s1.print(1, 0)) ; delete c1 ;
  s2 += 58 ;
  s2 += 2345 ;
  s2 += 1600 ;
  printf (", s2 = %s", c1 = s2.print(1, 0)) ; delete c1 ;
  s3 += 55 ;
  s3 += 1600 ;
  s3 += 58 ;
  printf (", and s3 = %s\n", c1 = s3.print(1, 0)) ; delete c1 ;
  s4 = s1 && s2 ;
  printf ("  then s4 = s1 && s2  is  %s\n", c1 = s4.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  s5 += 2345 ;
  if (s4 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }

  s4 = s2 && s3 ;
  printf ("       s4 = s2 && s3  is  %s\n", c1 = s4.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  s5 += 1600 ;
  if (s4 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }

  s4 = s1 && s3 ;
  printf ("       s4 = s1 && s3  is  %s\n", c1 = s4.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 55 ;
  s5 += 58 ;
  if (s4 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }
  s4 = s1 && s2 && s3 ;
  printf ("       s4 = s1 && s2 && s3  is  %s\n", c1 = s4.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  if (s4 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }

  // TEST 5: in-place Union operator

  printf ("\n TEST 5: in-place Union operator\n") ;
  s1.empty() ;
  s2.empty() ;
  s3.empty() ;
  s1 += 67 ;
  s1 += 451 ;
  printf ("  given s1 = %s", c1 = s1.print(1, 0)) ; delete c1 ;
  s2 += 77 ;
  s2 += 451 ;
  s2 += 10224 ;
  printf (", s2 = %s", c1 = s2.print(1,0)) ; delete c1 ;
  s3 += 16643 ;
  s3 += 1 ;
  s3 += 43 ;
  printf (", and s3 = %s\n", c1 = s3.print(1,0)) ; delete c1 ;
  s1 += s2 + s3 ;
  printf ("  then s1 += s2 + s3  yields s1 = %s\n", c1 = s1.print(1,0)) ;delete c1 ;
  s5.empty() ;
  s5 += 1 ;
  s5 += 43 ;
  s5 += 67 ;
  s5 += 451 ;
  s5 += 77 ;
  s5 += 10224 ;
  s5 += 16643 ;
  if (s1 == s5)
    printf ("  in-place set union operator OK\n") ;
  else {
    printf ("  ERROR: in-place set union operator FAILED\n") ;
    ++errs ;
  }

  // TEST 6: in-place Set intersection operator

  printf ("\n TEST 6: in-place intersection operator\n") ;
  s1.empty() ;
  s2.empty() ;
  s3.empty() ;
  s1 += 55 ;
  s1 += 58 ;
  s1 += 2345 ;
  s1 += 2901 ;
  printf ("  given s1 = %s", c1 = s1.print(1, 0)) ; delete c1 ;
  s2 += 58 ;
  s2 += 2345 ;
  s2 += 1600 ;
  printf (", s2 = %s", c1 = s2.print(1, 0)) ; delete c1 ;
  s3 += 55 ;
  s3 += 1600 ;
  s3 += 58 ;
  printf (", and s3 = %s\n", c1 = s3.print(1, 0)) ; delete c1 ;
  s1 &= s2 ;
  printf ("  then s1 &= s2 yields s1 = %s\n", c1 = s1.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  s5 += 2345 ;
  if (s1 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }

  s2 &= s3 ;
  printf ("       s2 &= s3 yields s2 = %s\n", c1 = s2.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  s5 += 1600 ;
  if (s2 != s5) {
    printf ("  ERROR: set intersection operator FAILED\n") ;
    ++errs ;
  }

  s1 &= s1 ;
  printf ("       s1 &= s1 yields s1 = %s\n", c1 = s1.print(1,0)) ; delete c1 ;
  s5.empty() ;
  s5 += 58 ;
  s5 += 2345 ;
  if (s1 != s5) {
    printf ("  ERROR: set self-intersection operator FAILED\n") ;
    ++errs ;
  }

  printf (" TEST 7: Iterating through a Set\n") ;
  int i ;
  printf ("  s1 contains {") ;
  for (SetIterator next(s1) ; i = next() ; ++next )
    printf (" %d", i) ;
  printf (" }\n") ;
}

void basic_speed_test()
{
  const int lim = 100000 ;   // repetition count for averaging performance...
  const int stride = 5 ;     // gap between element ids
  const int tests = lim/stride ;
  const int t0 = clock() ;   // start the clock running for this process
  Set s1 ;
  Set s2 ;
  Set s3 ;
  Set s4 ;
  Set s5 ;
  Set s6 ;
  char *c1 ;
  int errs = 0 ;
  int i ;

  printf ("\n\nPERFORMANCE BENCHMARKS ON BASIC SETS\n\n") ;
  // BM 1: simple addition to a single set
  
  int t1 = clock() ;
  for (i = 0; i < lim; i+=stride)
    s1 += i ;
  int usec = clock() - t1 ;
  printf ("BM-1: %d set insertions in %d usec = %d insertions/Sec\n", 
	  tests, usec, (int)((float)tests/usec*1000000)) ;
  
  // s1.analyze() ;

  // BM 2: false deletion from a single set
  
  t1 = clock() ;
  for (i = 0; i < lim; i+=stride)
    s1 -= (i + 1) ;
  usec = clock() - t1 ;
  printf ("BM-2: %d false deletions in %d usec = %d false deletions/Sec\n", 
	  tests, usec, (int)((float)tests/usec*1000000)) ;

  // BM 3: membership tests (absent)
  
  int bm3errs = 0 ;
  t1 = clock() ;
  for (i = 0; i < lim; i+=stride)
    if (s1[i + 1])
      ++bm3errs ;
  usec = clock() - t1 ;
  printf ("BM-3: %d absent membership tests in %d usec = %d memberships/Sec\n", 
	  tests, usec, (int)((float)tests/usec*1000000)) ;
  if (bm3errs)
    printf ("Warning: %d tests out of %d failed on BM-3\n", bm3errs, tests) ;

  // BM 4: membership tests (present)
  
  int bm4errs = 0 ;
  t1 = clock() ;
  for (i = 0; i < lim; i+=stride)
    if (!(s1[i]))
      ++bm4errs ;
  usec = clock() - t1 ;
  printf ("BM-4: %d present membership tests in %d usec = %d memberships/Sec\n", 
	  tests, usec, (int)((float)tests/usec*1000000)) ;
  if (bm4errs)
    printf ("Warning: %d tests out of %d failed on BM-4\n", bm4errs, tests) ;

  // BM 5: actual deletion from a single set
  
  t1 = clock() ;
  for (i = 0; i < lim; i+=stride)
    s1 -= i ;
  usec = clock() - t1 ;
  printf ("BM-5: %d actual deletions in %d usec = %d actual deletions/Sec\n", 
	  tests, usec, (int)((float)tests/usec*1000000)) ;

  if (!(s1.is_empty()))
    printf ("Warning: after deletions the set was not empty!\n") ;
}


void derived_regression()
{
  int ii;

  printf ("\nDERIVED SET REGRESSION TESTS\n\n") ;
  Universe<Tuple>* univ = new Universe<Tuple>(10,15) ;
  Tuple* t1 = univ->make_member(10) ;
  Tuple* t2 = univ->make_member(11) ;
  char *c1 ;

  printf (" Making members for s2 = {") ;
  for (ii = 0; ii < 50; ii++) {
    Tuple* t = univ->make_member() ;
    printf ("%d ", t->index()) ;
  }
  printf ("}\n") ;

  SetOf<Tuple> s1(univ) ;
  SetOf<Tuple> s2(univ) ;
  SetOf<Tuple> s3(univ) ;

  s1 += t1 ;
  s1 += t2 ;
  printf (" SetOf<Tuple> s1 = %s\n", c1 = s1.print(1,0)) ; 
  delete c1 ;
  
  for (ii = 1; ii < 50; ii++) {
    Tuple* t = univ->get_member(ii) ;
    if (t && (ii < 35))
      s2 += t ;
  }
  printf (" SetOf<Tuple> s2 = %s\n", c1 = s2.print(1,0)) ; 
  fflush(stdout) ;
  delete c1 ;

  s3 = s1 + s2 ;

  printf (" s3 = s1 + s2 = %s\n", c1 = s3.print(1,0)) ; 
  fflush(stdout) ;
  delete c1 ;

  s2 = ~s3 ;
  printf (" s2 = ~s3 = %s\n", c1 = s2.print(1,0)) ;
  fflush(stdout) ;
  delete c1 ;

  int t ;
  SetIteratorOf<Tuple> next(s2, SET_ACTUAL) ;
  printf (" Members of s2 = {") ;
  for ( ; t = next() ; ++next) {
    printf (" %d", t) ;
  }
  printf ("}\n") ;
  fflush(stdout) ;

  printf ("TESTING CO-UNIVERSES <Tuples, SetOfTuples>\n") ;
  fflush(stdout) ;

  Co_Universe<Tuple, TupleSet > defs (univ) ;

  Tuple* t20 = univ->get_member (20) ;

  TupleSet* sd = defs.get_codomain (t20) ;

  (*sd) += univ->get_member (24) ;
  (*sd) += univ->get_member (25) ;
  (*sd) += univ->get_member (27) ;

  printf (" installed tuples 24, 25, and 27 in SetOf<Tuple> sd\n") ;
  printf (" sd = %s\n", c1 = sd->print (1,0) ) ;
  fflush(stdout) ;
  delete c1 ;
  
  TupleSet* su = defs [t20] ;
  printf (" retreived codomain set su = %s\n", c1 = su->print (1,0) ) ;
  fflush(stdout) ;
  delete c1 ;

  delete univ ;
  return ;
}

void derived_speed_test()
{
  printf ("\nDERIVED SET PERFORMANCE BENCHMARKS\n\n") ;
}

// Surrogate error routines, only for testing...

void fatal(char *str)
{
  fprintf (stderr, "Fatal: %s\n", str) ;
}

void err(char *str)
{
  fprintf (stderr, "Error: %s\n", str) ;
}

int main(int argc, char* argv[])
{
  if (argc < 2) {
    printf ("Usage: test [-r|-b]\n") ;
    exit(0) ;
  }
  else if (!strcmp(argv[1], "-r")) {
    basic_regression() ;
    derived_regression() ;
  } else if (!strcmp(argv[1], "-b")) {
    basic_speed_test() ;
    derived_speed_test() ;
  }
  print_set_stats() ;
}
