//
// Class CCLParse for command-line parsing
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

public class CCLParse
{
  //Variables
  private int XNumOptMax;
  private int XIgnoreCase;
  private String XFileNameOpt;
  
  private String[] CL1;
  private String[] CL2;
  private String[] CL3;
  private int[] XHasParam;
  private String[] CL;

  private int CurParam;


  //****************************************************************************
  //"Initialization"
  //
  //Input:  NumOptMax   = Max Number of Options
  //        IgnoreCase  = 1 - Ignore case, 0 - Don't ignore case
  //        FileNameOpt = Name of file with options
  //                      (Options and parameters are on different lines
  //                      End is an empty line) 
  //
  //Return: 0  - Ok
  //        >0 - Error
  //

  public int Init(int NumOptMax,
                  int IgnoreCase,
                  String FileNameOpt)
  {
    int x;

    XNumOptMax=NumOptMax;
    XIgnoreCase=IgnoreCase;    
    XFileNameOpt=FileNameOpt;

    if (NumOptMax<1) return 1;

    if ((IgnoreCase!=0) && (IgnoreCase!=1)) return 1;

    CL1=new String [XNumOptMax];
    CL2=new String [XNumOptMax];
    CL3=new String [XNumOptMax];
    CL=new String[XNumOptMax];
    XHasParam=new int[XNumOptMax];

    for (x=0; x<XNumOptMax; x++) 
    {
       CL1[x]="";
       CL2[x]="";  
       CL3[x]="";
       CL[x]="";
       XHasParam[x]=0;
    }

    CurParam=0;

    return 0;
  }

  //****************************************************************************
  //"Add Option"
  //
  //Input:  Opt1     = Option 1
  //        Opt2     = Option 2 (or "") - alternative option
  //        Opt3     = Option 3 (or "") - alternative option
  //        Default  = Default Parameter
  //        HasParam = 0 - Option doesn't have parameter, 1 - it has
  //
  //Return: 0  - Ok
  //        >0 - Error
  //
  public int Add(String Opt1, String Opt2, String Opt3, String Default, int HasParam)
  {
    if (CurParam>XNumOptMax) return 1;

	CL1[CurParam]=Opt1;
  	CL2[CurParam]=Opt2;
	CL3[CurParam]=Opt3;

    if (XIgnoreCase==1)
    {
      CL1[CurParam]=CL1[CurParam].toLowerCase(); 
      CL2[CurParam]=CL2[CurParam].toLowerCase(); 
      CL3[CurParam]=CL3[CurParam].toLowerCase(); 
    }

    CL[CurParam]=Default;

    XHasParam[CurParam]=HasParam;

    CurParam++;

    return 0;
  }

  //****************************************************************************
  //"Parse Command Line"
  //
  //Return: 0  - Ok
  //        >0 - Error
  //
  public int Parse(String Args[])
  {
    int x, y;

    int NumArgs=Args.length;
    
    //Checking command line
    if (NumArgs>0)
    {
      int CurArg=1;

      do
      {
        String Args1=Args[CurArg-1];
        if (XIgnoreCase==1) Args1=Args1.toLowerCase();
    
        y=0;
        for (x=0; x<CurParam; x++)
        {
          if      (Args1.equals(CL1[x])) {y=1; break;}
          else if (Args1.equals(CL2[x])) {y=1; break;}
          else if (Args1.equals(CL3[x])) {y=1; break;}
        } 

        if (y==1)
        {
          if (XHasParam[x]==0) CL[x]="1";
          else
          {
            CurArg++;
            if (CurArg>NumArgs) return 2; // Option without parameter
            CL[x]=Args[CurArg-1];
          }
        }
        else return 1; // Unknown option
  
        CurArg++;

      }
      while (CurArg<=NumArgs);
    }

    //Checking file with parameters
    //Don't use ios::nocreate in Linux!
    BufferedReader fin=null;

    try
    {
      fin=new BufferedReader(new FileReader(XFileNameOpt));
    }
    catch (IOException e2)
    {
      return 0;
    }

    String str1="";
    String str2="";

    try
    {
      while ((str1=fin.readLine())!=null)
      {
        if (XIgnoreCase==1) str1=str1.toLowerCase();

        y=0;
        for (x=0; x<CurParam; x++)
        {
          if      (str1.equals(CL1[x])) {y=1; break;}
          else if (str1.equals(CL2[x])) {y=1; break;}
          else if (str1.equals(CL3[x])) {y=1; break;}
        } 

        if (y==1)
        {
          if (XHasParam[x]==0) CL[x]="1";
          else
          {
            str2=fin.readLine();
            if (str2==null) return 3; //Option without parameter in file
          
            CL[x]=str2;
          }
        }
        else return 4; //Unknown option in file    
      }
    }
    catch (Exception e1)
    {
      try
      {
        fin.close();
      }
      catch (Exception e2)
      {}
      return 0;
    }

    try
    {
      fin.close();
    }
    catch (Exception e3)
    {}
      
    return 0;
  }

  //****************************************************************************
  //"Get String Parameter"
  //
  public String GetParamStr(int NumParam)
  {   
    if ((NumParam<0) || (NumParam>=CurParam)) return null;

    return CL[NumParam];
  }

  //****************************************************************************
  //"Get Long Parameter"
  //
  public long GetParamLong(int NumParam)
  {
    if ((NumParam<0) || (NumParam>=CurParam)) return 0;

    long ln=0;
    try
    {
      ln=Long.valueOf(CL[NumParam]).longValue();
    }
    catch (NumberFormatException nfe1)
    {}
    
    return ln;
  }

  //****************************************************************************
  //"Get Int Parameter"
  //
  public int GetParamInt(int NumParam)
  {
    if ((NumParam<0) || (NumParam>=CurParam)) return 0;

    int in=0;
    try
    {
      in=Integer.valueOf(CL[NumParam]).intValue();
    }
    catch (NumberFormatException nfe1)
    {}
    
    return in;
  }
}
